from syft.federated.train_config import TrainConfig
from syft.federated import federated_client
from syft.federated.federated_client import FederatedClient

__all__ = ["TrainConfig", "federated_client"]
